/*
 * Program to control MB1501 PLLs
 * for the TKR-901 -- Kenwood 900 MHz repeater
 *
 * Neal Reasoner, KB5ERY
 * Copyright 2014
 *
 * Changes made to original program to reflect the new 40 pin header
 * Re asignment of GPIO pins.
 * Graham Warrington VE3WGW
 * gcc pll.c -o pll -Wall; strip pll
 */

#define REF_OSC 12.8		// 12.8 MHz referece oscillator
#define CHAN_SPACING 0.0125	// 12.5 kHz channel spacing
#define IF_FREQ_OFFSET -39.0375	// IF frequency (for RX PLL freq adjustment)


// NOTE:  PULL-DOWN SETUP ASSUMES ALL GPIOs USED ARE IN RANGE 0-31
//GPIO number
#define TX_CLK 5		// Connector P1 - pin 29
#define TX_DATA 6		// Connector P1 - pin 31
#define TX_ENA 13		// Connector P1 - pin 33

#define RX_CLK 17		// Connector P1 - pin 11
#define RX_DATA 27		// Connector P1 - pin 13
#define RX_ENA 22		// Connector P1 - pin 15
/* P1 connector pinout - 40 pin ******************************************
*		P1-1	3.3v			P1-2	5.0v
*		P1-3	GPIO 2 SDA1		P1-4	5.0v
*		P1-5	GPIO 3 SCL1		P1-6	Ground
*		P1-7	GPIO 4			P1-8	GPIO14	TXD0
*		P1-9	Ground			P1-10	GPIO15	RXD0
*		P1-11	GPIO 17			P1-12	GPIO18
*		P1-13	GPIO 27			P1-14	Ground
*		P1-15	GPIO 22			P1-16	GPIO23
*		P1-17	3.3v			P1-18	GPIO24
*		P1-19	GPIO 10 SPIMOSI	        P1-20	Ground
*		P1-21	GPIO 9 SPI_MISO	        P1-22	GPIO25
*		P1-23	GPIO 11 SPISCLK	        P1-24	GPIO8	SPI_CE0_n
*		P1-25	GND		        P1-26	GPIO7	SPI_CE1_n
*               P1-27   ID SD                   P1-28   ID SC
*               P1-29   GPIO 5                  P1-30   Ground
*               P1-31   GPIO 6                  P1-32   GPIO 12
*               P1 33   GPIO 13                 P1-34   Ground 
*               P1-35   GPIO1 9 PCM FS          P1-36   GPIO 16
*               P1-37   GPIO 26                 P1-38   GPIO 20 PCM DIN
*               P1-39 	Ground                  P1-40   GPIO 21 PCM DOUT
*		                               
*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>
#include <string.h>


// GPIO macros for Raspberry Pi
#include <fcntl.h>
#include <sys/mman.h>

#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIO controller */
#define BLOCK_SIZE (4*1024)

// GPIO setup macros. Always use INP_GPIO(x) before using OUT_GPIO(x) or SET_GPIO_ALT(x,y)
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

#define GPIO_SET(g) (g<32 ? (*(gpio+7)=(1<<g)) : (*(gpio+8)=(1<<(g-32))))  // sets   bits which are 1 ignores bits which are 0
#define GPIO_CLR(g) (g<32 ? (*(gpio+10)=(1<<g)) : (*(gpio+11)=(1<<(g-32)))) // clears bits which are 1 ignores bits which are 0
#define READ_GPIO(g) (g<32 ? ((*(gpio+13)&(1<<g))?1:0) : ((*(gpio+14)&(1<<(g-32)))?1:0)) // read pin values

#define GPPUD (gpio+37)
#define GPPUDCLKA (gpio+38)		// GPIOs 0-31
#define GPPUDCLKB (gpio+39)		// GPIOs 32-53

// Function declarations
void setup_gpio();
void print_usage(int argc, char *argv[]);
void set_pll(float freq, int CLK, int DATA, int ENA, int verbose);

// global for GPIO access
volatile unsigned *gpio;

#define PRE(sw)	(sw==0 ? 128 : 64)


int main(int argc, char *argv[])
{
	float rxfreq = 0;
	float txfreq = 0;
	int verbose = 0;
	int bad_option = 0;
	int i;
	struct timespec delay;
	
	printf("%s -- MB1501 PLL control for TKR-901\r\n\r\n", argv[0]);
	if (argc<2) {
		print_usage(argc, argv);
		exit(0);
	}
	for (i=1; i<argc; i++) {
		if (!strcmp(argv[i], "-r")) {
			if (!sscanf(argv[++i], "%f", &rxfreq)) {
				printf("Receive frequency formatted incorrectly!\r\n\r\n");
				print_usage(argc, argv);
				exit(0);
			}
		} else if (!strcmp(argv[i], "-t")) {
			if (!sscanf(argv[++i], "%f", &txfreq)) {
				printf("Transmit frequency formatted incorrectly!\r\n\r\n");
				print_usage(argc, argv);
				exit(0);
			}
		} else if (!strcmp(argv[i], "-v")) {
			verbose = 1;
		} else {
			printf("Option %s not recognized.\r\n", argv[i]);
			bad_option = 1;
		}
	}
	if (bad_option) print_usage(argc, argv);

	setup_gpio();

	// Set our GPIO pins to pull-down
	// NOTE:  THIS ASSUMES ALL GPIOs USED ARE IN RANGE 0-31
	*GPPUDCLKA = 0;
	*GPPUD = 1;		// 0 off, 1 pull-down, 2 pull-up
	delay.tv_sec = 0;
	delay.tv_nsec = (long)(1000);		// delay in nanoseconds
	nanosleep(&delay, 0);
	*GPPUDCLKA = (1<<TX_CLK) | (1<<TX_DATA) | (1<<TX_ENA) | (1<<RX_CLK) | (1<<RX_DATA) | (1<<RX_ENA);
	nanosleep(&delay, 0);
	*GPPUDCLKA = 0;

	// Set our GPIO pins as inputs -- must use INP_GPIO before we can use OUT_GPIO
	INP_GPIO(TX_CLK);
	INP_GPIO(TX_DATA);
	INP_GPIO(TX_ENA);

	INP_GPIO(RX_CLK);
	INP_GPIO(RX_DATA);
	INP_GPIO(RX_ENA);

	if (rxfreq > 0) {
		printf("\r\nSetting receive frequency to %.4f MHz ", rxfreq);
		rxfreq += (IF_FREQ_OFFSET);		// adjust for receiver IF frequency
		printf("with VCO freq of %.4f MHz\r\n", rxfreq);
		set_pll(rxfreq, RX_CLK, RX_DATA, RX_ENA, verbose);
	}

	if (txfreq > 0) {
		printf("\r\nSetting transmit frequency to %.4f MHz\r\n", txfreq);
		set_pll(txfreq, TX_CLK, TX_DATA, TX_ENA, verbose);
	}

	return(0);
}



void set_pll(float freq, int CLK, int DATA, int ENA, int verbose)
{
	int vcodiv, refdiv, sw, vcoswallow, i;
	struct timespec delay;

	delay.tv_sec = 0;
	delay.tv_nsec = (long)(10000);		// delay in nanoseconds, CLK period is 2x this number

	sw = 0;	// prescaler setting:  0=128, 1=64

	vcodiv = (int)(freq / CHAN_SPACING / PRE(sw));
	if ( (vcodiv < 16) || (vcodiv > 2047) ) {
		printf("VCO divide value out of range!\r\n");
		return;
	}
	vcoswallow = ((int)(freq / CHAN_SPACING)) - vcodiv*PRE(sw);
	if ( (vcoswallow < 0) || (vcoswallow >= PRE(sw)) ) {
		// this can never actually happen unless there's a bug
		printf("VCO swallow counter out of range!\r\n");
		return;
	}

	printf("Setting VCO to divide by %d = %d * %d + %d\r\n", (int)(freq / CHAN_SPACING), vcodiv, PRE(sw), vcoswallow);

	vcodiv <<= 7;
	vcodiv += vcoswallow;

	vcodiv <<= 1;
	vcodiv &= ~1;			// always clear C (control bit) to program VCO divider

	refdiv = (int)(REF_OSC / CHAN_SPACING);
	if ( (refdiv > 16383) || (refdiv < 8) ) {
		printf("Reference oscillator divide value out of range!\r\n");
		return;
	}
	printf("Setting reference to divide by %d\r\n", refdiv);
	refdiv <<= 1;
	refdiv |= sw<<15;		// set SW (prescaler) to 64 if high, 128 if low
	refdiv |= 1;			// always set C (control bit) to program ref divider


	// inputs are pulled down, so start with all bits clear
	GPIO_CLR(CLK);
	GPIO_CLR(DATA);
	GPIO_CLR(ENA);
	// set TX GPIO pins to be outputs
	OUT_GPIO(CLK);
	OUT_GPIO(DATA);
	OUT_GPIO(ENA);

	nanosleep(&delay, 0);

	// Set reference divider
	for (i=0; i<16; i++) {
		if (refdiv & 0x8000) {
			GPIO_SET(DATA);
			if (verbose) printf("1");
		} else {
			GPIO_CLR(DATA);
			if (verbose) printf("0");
		}
		nanosleep(&delay, 0);
		GPIO_SET(CLK);
		nanosleep(&delay, 0);
		GPIO_CLR(CLK);
		refdiv <<= 1;
	}
	nanosleep(&delay, 0);
	GPIO_SET(ENA);
	nanosleep(&delay, 0);
	GPIO_CLR(ENA);
	nanosleep(&delay, 0);
	if (verbose) printf("\r\n");

	// Set VCO divider
	for (i=0; i<19; i++) {
		if (vcodiv & 0x40000) {
			GPIO_SET(DATA);
			if (verbose) printf("1");
		} else {
			GPIO_CLR(DATA);
			if (verbose) printf("0");
		}
		nanosleep(&delay, 0);
		GPIO_SET(CLK);
		nanosleep(&delay, 0);
		GPIO_CLR(CLK);
		vcodiv <<= 1;
	}
	nanosleep(&delay, 0);
	GPIO_SET(ENA);
	nanosleep(&delay, 0);
	GPIO_CLR(ENA);
	nanosleep(&delay, 0);
	if (verbose) printf("\r\n");

	// INP_GPIO(CLK);
	// INP_GPIO(DATA);
	// INP_GPIO(ENA);

	return;
}



void print_usage(int argc, char *argv[])
{
	printf("Usage:  %s [-r value] [-t value]\r\n\r\n", argv[0]);
	printf("\t-r\tset the receive frequency\r\n");
	printf("\t-t\tset the transmit frequency\r\n");
	printf("\twhere value is the frequency in megahertz\r\n\r\n");
	printf("\tfor example:  %s -r 902.3675 -t 927.3675\r\n\r\n", argv[0]);
}


// Set up a memory regions to access GPIO
void setup_gpio()
{
	int  mem_fd;
	void *gpio_map;

	/* open /dev/mem */
	if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
		printf("can't open /dev/mem \n");
		exit(-1);
	}

	/* mmap GPIO */
	gpio_map = mmap(
			NULL,             //Any adddress in our space will do
			BLOCK_SIZE,       //Map length
			PROT_READ|PROT_WRITE,// Enable reading & writting to mapped memory
			MAP_SHARED,       //Shared with other processes
			mem_fd,           //File to map
			GPIO_BASE         //Offset to GPIO peripheral
			);

	close(mem_fd); //No need to keep mem_fd open after mmap

	if (gpio_map == MAP_FAILED) {
		printf("mmap error %d\n", (int)gpio_map);//errno also set!
		exit(-1);
	}

	// Always use volatile pointer!
	gpio = (volatile unsigned *)gpio_map;
} // setup_gpio
